#
#   Copyright (c) 2014-2015, 2018 eGauge Systems LLC
# 	1644 Conestoga St, Suite 2
# 	Boulder, CO 80301
# 	voice: 720-545-9767
# 	email: davidm@egauge.net
#
#   All rights reserved.
#
#   This code is the property of eGauge Systems LLC and may not be
#   copied, modified, or disclosed without any prior and written
#   permission from eGauge Systems LLC.
#
from django.contrib.auth import models as auth_models
from django.contrib.auth.models import Permission
from django.contrib.contenttypes.models import ContentType
from django.db.models.signals import post_migrate

from epic.apps import EPIC_App_Config as app

VIEW_NAME = "view"
EDIT_NAME = "edit"

VIEW = "%s.%s" % (app.name, VIEW_NAME)
EDIT = "%s.%s" % (app.name, EDIT_NAME)

VIEW_PERM = None
EDIT_PERM = None


def add_custom_permissions(sender, **kwargs):
    # pylint: disable=unused-argument, global-statement
    global VIEW_PERM, EDIT_PERM

    ct, _ = ContentType.objects.get_or_create(app_label=app.name, model="")
    p, _ = Permission.objects.get_or_create(
        codename=VIEW_NAME, name="can view", content_type=ct
    )
    VIEW_PERM = p
    p, _ = Permission.objects.get_or_create(
        codename=EDIT_NAME, name="can edit/delete", content_type=ct
    )
    EDIT_PERM = p


post_migrate.connect(add_custom_permissions, sender=auth_models)



#
#   Copyright (c) 2014-2016, 2018, 2021-2022 eGauge Systems LLC
# 	1644 Conestoga St, Suite 2
# 	Boulder, CO 80301
# 	voice: 720-545-9767
# 	email: davidm@egauge.net
#
#   All rights reserved.
#
#   This code is the property of eGauge Systems LLC and may not be
#   copied, modified, or disclosed without any prior and written
#   permission from eGauge Systems LLC.
#
from django.urls import include, re_path

from epic import exports, views

urlpatterns = [
    re_path(r"^$", views.epic_index, name="epic_index"),
    re_path(r"^search-results$", views.search_results, name="search_results"),
    re_path(r"^api/", include("epic.api.urls")),
    re_path(r"^part/?$", views.part_list, name="part_list"),
    re_path(r"^part/add$", views.part_add, name="part_add"),
    re_path(r"^part/info$", views.part_info, name="part_info"),
    re_path(r"^part/(?P<pk>\d+)/?$", views.part_detail, name="part_detail"),
    re_path(r"^part/(?P<pk>\d+)/edit$", views.part_edit, name="part_edit"),
    re_path(
        r"^part/(?P<pk>\d+)/delete$", views.part_delete, name="part_delete"
    ),
    re_path(
        r"^part/(?P<pk>\d+)/bom/?$",
        views.part_bom_detail,
        name="part_bom_detail",
    ),
    re_path(
        r"^part/(?P<pk>\d+)/bom/edit$",
        views.part_bom_edit,
        name="part_bom_edit",
    ),
    re_path(
        r"^part/(?P<pk>\d+)/bom/export$",
        exports.part_bom_export,
        name="part_bom_export",
    ),
    re_path(
        r"^part/(?P<pk>\d+)/bom/compare/?$",
        views.part_bom_compare,
        name="part_bom_compare",
    ),
    re_path(
        r"^part/(?P<pk>\d+)/bom/compare/(?P<other_pk>\d+)/?$",
        views.part_bom_compare_qty,
        name="part_bom_compare_qty",
    ),
    re_path(
        r"^part/(?P<pk>\d+)/bom/compare/(?P<other_pk>\d+)/refdes$",
        views.part_bom_compare_refdes,
        name="part_bom_compare_refdes",
    ),
    re_path(
        r"^part/datasheet/?$", views.datasheet_list, name="datasheet_list"
    ),
    re_path(
        r"^part/datasheet/add$", views.datasheet_add, name="datasheet_add"
    ),
    re_path(
        r"^part/datasheet/add/(?P<pk>\d+)$",
        views.datasheet_add_part,
        name="datasheet_add_part",
    ),
    re_path(
        r"^part/datasheet/(?P<pk>\d+)/?$",
        views.datasheet_detail,
        name="datasheet_detail",
    ),
    re_path(
        r"^part/datasheet/(?P<pk>\d+)/edit$",
        views.datasheet_edit,
        name="datasheet_edit",
    ),
    re_path(
        r"^part/datasheet/(?P<pk>\d+)/delete$",
        views.datasheet_delete,
        name="datasheet_delete",
    ),
    re_path(r"^vendor/?$", views.vendor_list, name="vendor_list"),
    re_path(r"^vendor/add$", views.vendor_add, name="vendor_add"),
    re_path(
        r"^vendor/(?P<pk>\d+)/?$", views.vendor_detail, name="vendor_detail"
    ),
    re_path(
        r"^vendor/(?P<pk>\d+)/edit$", views.vendor_edit, name="vendor_edit"
    ),
    re_path(
        r"^vendor/(?P<pk>\d+)/delete$",
        views.vendor_delete,
        name="vendor_delete",
    ),
    re_path(r"^warehouse/?$", views.warehouse_list, name="warehouse_list"),
    re_path(
        r"^warehouse/stock$",
        views.warehouse_stock_all,
        name="warehouse_stock_all",
    ),
    re_path(
        r"^warehouse/stock/export$",
        exports.warehouse_stock_all_export,
        name="warehouse_stock_all_export",
    ),
    re_path(r"^warehouse/add$", views.warehouse_add, name="warehouse_add"),
    re_path(
        r"^warehouse/(?P<pk>\d+)/?$",
        views.warehouse_detail,
        name="warehouse_detail",
    ),
    # this is simply an alias since the warehouse_detail already lists
    # the inventories associated with that warehouse:
    re_path(
        r"^warehouse/(?P<pk>\d+)/inventory/?$",
        views.warehouse_inventory,
        name="warehouse_inventory",
    ),
    re_path(
        r"^warehouse/(?P<pk>\d+)/edit$",
        views.warehouse_edit,
        name="warehouse_edit",
    ),
    re_path(
        r"^warehouse/(?P<pk>\d+)/add-shipment$",
        views.warehouse_add_shipment,
        name="warehouse_add_shipment",
    ),
    re_path(
        r"^warehouse/(?P<pk>\d+)/add-inventory$",
        views.warehouse_add_inventory,
        name="warehouse_add_inventory",
    ),
    re_path(
        r"^warehouse/(?P<pk>\d+)/stock$",
        views.warehouse_stock,
        name="warehouse_stock",
    ),
    re_path(
        r"^warehouse/(?P<pk>\d+)/stock/export$",
        exports.warehouse_stock_export,
        name="warehouse_stock_export",
    ),
    re_path(
        r"^warehouse/(?P<pk>\d+)/delete$",
        views.warehouse_delete,
        name="warehouse_delete",
    ),
    re_path(
        r"^warehouse/(?P<warehouse>\d+)/inventory/(?P<pk>\d+)/?$",
        views.warehouse_inv_detail,
        name="warehouse_inventory_detail",
    ),
    re_path(
        r"^warehouse/(?P<warehouse>\d+)/inventory/(?P<pk>\d+)/export$",
        exports.warehouse_inv_export,
        name="warehouse_inventory_export",
    ),
    re_path(
        r"^warehouse/(?P<warehouse>\d+)/inventory/(?P<pk>\d+)/edit$",
        views.inventory_edit,
        name="warehouse_inventory_edit",
    ),
    re_path(
        r"^warehouse/(?P<warehouse>\d+)/inventory/(?P<pk>\d+)/delete$",
        views.inventory_delete,
        name="warehouse_inventory_delete",
    ),
    re_path(r"^order/?$", views.order_list, name="order_list"),
    re_path(r"^order/add$", views.order_add, name="order_add"),
    re_path(r"^order/(?P<pk>\d+)/?$", views.order_detail, name="order_detail"),
    re_path(
        r"^order/(?P<pk>\d+)/export$",
        exports.order_export,
        name="order_export",
    ),
    re_path(r"^order/(?P<pk>\d+)/edit$", views.order_edit, name="order_edit"),
    re_path(
        r"^order/(?P<pk>\d+)/check-stock$",
        views.order_check_stock,
        name="order_check_stock",
    ),
    re_path(
        r"^order/(?P<pk>\d+)/add-shipment$",
        views.order_add_shipment,
        name="order_add_shipment",
    ),
    re_path(
        r"^order/(?P<pk>\d+)/delete$", views.order_delete, name="order_delete"
    ),
    re_path(r"^ship/?$", views.ship_list, name="ship_list"),
    re_path(r"^ship/add$", views.ship_add, name="ship_add"),
    re_path(r"^ship/(?P<pk>\d+)/?$", views.ship_detail, name="ship_detail"),
    re_path(
        r"^ship/(?P<pk>\d+)/export$", exports.ship_export, name="ship_export"
    ),
    re_path(r"^ship/(?P<pk>\d+)/edit$", views.ship_edit, name="ship_edit"),
    re_path(
        r"^ship/(?P<pk>\d+)/delete$", views.ship_delete, name="ship_delete"
    ),
    re_path(
        r"^dal/part/?$", views.Part_Autocomplete.as_view(), name="part-dal"
    ),
    re_path(
        r"^dal/assy/?$",
        views.Assembly_Autocomplete.as_view(),
        name="assembly-dal",
    ),
    re_path(
        r"^dal/order/?$", views.Order_Autocomplete.as_view(), name="order-dal"
    ),
    re_path(r"^dal/mfg/?$", views.Mfg_Autocomplete.as_view(), name="mfg-dal"),
    re_path(
        r"^dal/footprint/?$",
        views.Footprint_Autocomplete.as_view(),
        name="footprint-dal",
    ),
]



class Error(Exception):
    """Used for exceptions raised by the kicad package."""


class UnknownFlavorsError(Error):
    """Receives two argumets: a list of unknown flavors and a list of
    detected flavors."""



# Generated by Django 3.2.4 on 2021-11-08 17:18

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('epic', '0009_auto_20191003_2243'),
    ]

    operations = [
        migrations.AlterField(
            model_name='assembly_item',
            name='comp',
            field=models.ForeignKey(help_text='The part number of the component of this item.', on_delete=django.db.models.deletion.PROTECT, related_name='assembly_item_comp', to='epic.part', verbose_name='Component Part #'),
        ),
        migrations.AlterField(
            model_name='delta',
            name='part',
            field=models.ForeignKey(help_text='The part whose quantity gets adjusted.', on_delete=django.db.models.deletion.PROTECT, to='epic.part', verbose_name='Part #'),
        ),
        migrations.AlterField(
            model_name='delta',
            name='warehouse',
            field=models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, to='epic.warehouse'),
        ),
        migrations.AlterField(
            model_name='line_item',
            name='part',
            field=models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, to='epic.part', verbose_name='Part #'),
        ),
        migrations.AlterField(
            model_name='order',
            name='vendor',
            field=models.ForeignKey(help_text='The name of the vendor (distributor) where the order was placed.', on_delete=django.db.models.deletion.PROTECT, to='epic.vendor'),
        ),
        migrations.AlterField(
            model_name='part',
            name='datasheet',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.PROTECT, to='epic.datasheet'),
        ),
        migrations.AlterField(
            model_name='part',
            name='substitute',
            field=models.ForeignKey(blank=True, help_text='List of other parts that are (identical) substitutes for this part.', null=True, on_delete=django.db.models.deletion.PROTECT, to='epic.part', verbose_name='Substitutes'),
        ),
        migrations.AlterField(
            model_name='shipment',
            name='from_warehouse',
            field=models.ForeignKey(blank=True, help_text='For an inter-warehouse shipment, the warehouse the shipment originates from.', null=True, on_delete=django.db.models.deletion.PROTECT, to='epic.warehouse'),
        ),
        migrations.AlterField(
            model_name='shipment',
            name='ordr',
            field=models.ForeignKey(blank=True, help_text='For an order shipment, the order that resulted in this shipment.', null=True, on_delete=django.db.models.deletion.PROTECT, to='epic.order', verbose_name='Order #'),
        ),
        migrations.AlterField(
            model_name='stock_cache_item',
            name='part',
            field=models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, to='epic.part', verbose_name='Part #'),
        ),
        migrations.AlterField(
            model_name='stock_cache_item',
            name='warehouse',
            field=models.ForeignKey(help_text='The warehouse stocking this part.', on_delete=django.db.models.deletion.PROTECT, to='epic.warehouse', verbose_name='Warehouse'),
        ),
        migrations.AlterField(
            model_name='transaction',
            name='warehouse',
            field=models.ForeignKey(help_text='The (destination) warehouse this transaction applies to.', on_delete=django.db.models.deletion.PROTECT, to='epic.warehouse', verbose_name='Warehouse'),
        ),
    ]



#
#   Copyright (c) 2019 eGauge Systems LLC
# 	1644 Conestoga St, Suite 2
# 	Boulder, CO 80301
# 	voice: 720-545-9767
# 	email: davidm@egauge.net
#
#   All rights reserved.
#
#   This code is the property of eGauge Systems LLC and may not be
#   copied, modified, or disclosed without any prior and written
#   permission from eGauge Systems LLC.
#
"""Helper class for managing stock cache records."""

from django.utils import timezone

from epic.models import Delta, Part_Inventory, Stock_Cache_Entry


def _cache_ts(ts):
    """Given the point in time `ts', return the timestamp of the nearest
    cache record that is not younger than `ts'.

    """
    if ts is None:
        ts = timezone.now()
    # round down to beginning of day:
    return ts.replace(day=1, hour=0, minute=0, second=0, microsecond=0)


class Stock_Cache:
    @classmethod
    def get(cls, as_of_ts):
        """Get most recent part inventory from the cache.  The inventory is
        for a time that is not newer than `as_of_ts'.  If `as_of_ts'
        is None, the most recent part inventory is returned.
        Generally, the returned inventory may be older than `as_of_ts'
        since cache entries are created only at a certain granularity
        (e.g., at most one cache per day).  Consult the timestamp of
        the inventory to find out the actual time the returned
        inventory is for.

        """
        # find most recent stock cache entry, if any:
        qs = Stock_Cache_Entry.objects.all()
        if as_of_ts is not None:
            qs = qs.filter(ts__lte=as_of_ts)
        if qs.exists():
            stock_cache_entry = qs[0]
            inv = stock_cache_entry.inventory()
            # print('  STOCK_CACHE: found %s for %s' % (inv.ts, as_of_ts))
        else:
            inv = Part_Inventory()

        cache_ts = _cache_ts(as_of_ts)
        if inv.ts is None or inv.ts < cache_ts:
            # apply all deltas for transactions between inv.ts and cache_ts:
            qs = Delta.objects.order_by("txtn__ts")
            if inv.ts is not None:
                qs = qs.filter(txtn__ts__gt=inv.ts)
            qs = qs.filter(txtn__ts__lte=cache_ts)
            for d in qs:
                inv.apply_delta(d)

            # create a new stock cache entry for cache_ts:
            inv.ts = cache_ts
            new_stock_cache_entry = Stock_Cache_Entry()
            new_stock_cache_entry.save(inventory=inv)
            # print('  STOCK_CACHE: created %s for %s' % (inv.ts, as_of_ts))
        return inv



# Generated by Django 2.0.2 on 2018-02-03 05:53

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('epic', '0007_auto_20170510_0436'),
    ]

    operations = [
        migrations.AlterField(
            model_name='part',
            name='status',
            field=models.IntegerField(choices=[(0, 'preview'), (1, 'active'), (2, 'preferred'), (3, 'deprecated'), (4, 'obsolete')], default=0, help_text="The life-time status of this part. Parts marked `preview' and `obsolete' are not considered orderable.", verbose_name='Life-time Status'),
        ),
        migrations.AlterField(
            model_name='vendor_part',
            name='status',
            field=models.IntegerField(choices=[(0, 'preview'), (1, 'active'), (2, 'preferred'), (3, 'deprecated'), (4, 'obsolete')], default=1, help_text='The life-time status of this vendor part.', verbose_name='Life-time Status'),
        ),
    ]



#
#   Copyright (c) 2021-2022, 2025 eGauge Systems LLC
# 	4805 Sterling Dr, Suite 1
# 	Boulder, CO 80301
# 	voice: 720-545-9767
# 	email: davidm@egauge.net
#
#   All rights reserved.
#
#   This code is the property of eGauge Systems LLC and may not be
#   copied, modified, or disclosed without any prior and written
#   permission from eGauge Systems LLC.
#
"""KiCad EPIC scripts."""

import argparse
import logging
import sys
import traceback
from argparse import RawDescriptionHelpFormatter as RawDescHelpFormatter
from pathlib import Path

from epic.base import format_part_number

from . import schematics
from .bom import BOM
from .csv import CSVBOM
from .epic_client import EPICAPIClient
from .error import Error, UnknownFlavorsError
from .vendors_cache import VendorsCache


def epic_bom(epic_url, auth):
    """Command-line interface to translate a KiCAD XML file to a BOM."""
    parser = argparse.ArgumentParser(
        description=__doc__, formatter_class=RawDescHelpFormatter
    )
    parser.add_argument(
        "input-file",
        help='Name of XML file generated by KiCad\'s "Generate BOM" command.',
    )
    parser.add_argument(
        "output-file",
        help="Name of output file "
        "that will contain the bill-of-materials (BOM) in "
        "a comma-separated-values (CSV) file.",
    )
    parser.add_argument(
        "-d",
        "--debug",
        action="store_const",
        const=logging.DEBUG,
        dest="log_level",
        default=logging.INFO,
        help="Show debug output.",
    )
    parser.add_argument(
        "-f",
        "--force-overwrite",
        action="store_true",
        help="Force overwriting of existing EPIC BOM "
        "(assembly) part even if the part appears to "
        "have been modified manually.",
    )
    parser.add_argument(
        "--flavor",
        action="append",
        help="Select PCB flavor for which to generate BOM for."
        "  This option may be specified multiple times.",
    )
    parser.add_argument(
        "-q",
        "--quiet",
        action="store_true",
        help="Silence all output except for errors.",
    )
    parser.add_argument(
        "-p",
        "--prefer-vendor",
        nargs="+",
        help="Like --with-vendor-pn, but only output vendor "
        "part numbers for the specified vendor(s).",
    )
    parser.add_argument(
        "-V",
        "--with-vendor-pn",
        help="Include vendor part numbers in the CSV file.",
        action="store_true",
    )
    args = parser.parse_args()

    if args.quiet:
        args.log_level = logging.ERROR

    logging.basicConfig()
    log = logging.getLogger()
    log.setLevel(args.log_level)

    epic_api = EPICAPIClient(epic_url, auth)
    input_file = getattr(args, "input-file")
    try:
        bom = BOM(
            epic_api,
            input_file,
            "eGauge",
            flavors=args.flavor,
            toolname="epic-bom",
        )
        assy, created = bom.save_as_epic_assembly(
            force_update=args.force_overwrite
        )
        if not args.quiet:
            if bom.flavors:
                if len(bom.flavors) > 1:
                    detected_flavors_str = "flavors: "
                else:
                    detected_flavors_str = "flavor: "
                detected_flavors_str += ", ".join(sorted(bom.flavors))
            else:
                detected_flavors_str = "no flavors"
            log.info("this schematic defines %s", detected_flavors_str)

            print(
                "%s part %s (%s %s) with %d components."
                % (
                    "Created" if created else "Updated",
                    format_part_number(assy.id),
                    assy.mfg,
                    assy.mfg_pn,
                    len(bom.comps),
                )
            )

        if args.with_vendor_pn:
            # include all vendors that have relevant parts:
            vendors = []
        elif args.prefer_vendor:
            vendors = []
            for name in args.prefer_vendor:
                v = VendorsCache.find(name, relaxed_match=True)
                if not v:
                    raise Error(f"vendor `{name}' unknown")
                vendors.append(v)
        else:
            vendors = None

        CSVBOM(bom, vendors).save(getattr(args, "output-file"))

    except UnknownFlavorsError as e:
        log.error(
            "Flavor(s) %s are unknown; detected flavors: %s",
            ",".join(e.args[0]),
            ",".join(e.args[1]),
        )
        sys.exit(1)
    except Error as e:
        log.error(e.args)
        traceback.print_exc()
        sys.exit(1)


def epic_footprints(epic_url, auth):
    """Command-line interface to update KiCad schematic files with the
    footprints from EPIC.

    """
    parser = argparse.ArgumentParser(
        description=__doc__, formatter_class=RawDescHelpFormatter
    )
    parser.add_argument(
        "input-file", type=Path, help="Name of KiCad schematic file."
    )
    parser.add_argument(
        "-d",
        "--debug",
        action="store_const",
        const=logging.DEBUG,
        dest="log_level",
        default=logging.INFO,
        help="Show debug output.",
    )
    parser.add_argument(
        "-q",
        "--quiet",
        action="store_const",
        const=logging.ERROR,
        dest="log_level",
        help="Silence all output except for errors.",
    )
    args = parser.parse_args()

    logging.basicConfig()
    log = logging.getLogger()
    log.setLevel(args.log_level)

    epic_api = EPICAPIClient(epic_url, auth)

    path = getattr(args, "input-file")
    if path.suffix == ".sch":
        log.error("Cannot update old (pre v6) KiCad schematics files.")
        sys.exit(1)

    schematics.update_footprints(epic_api, path)



# -*- coding: utf-8 -*-
# Generated by Django 1.11.1 on 2017-05-09 04:22
from __future__ import unicode_literals

from decimal import Decimal
import django.core.validators
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('epic', '0005_alter_datasheet_ds_file'),
    ]

    operations = [
        migrations.AddField(
            model_name='part',
            name='notes',
            field=models.TextField(blank=True),
        ),
        migrations.AlterField(
            model_name='assembly_item',
            name='assy',
            field=models.ForeignKey(help_text='The part number of the assembly this item belongs to.', on_delete=django.db.models.deletion.CASCADE, related_name='assembly_item_part', to='epic.Part', verbose_name='Assembly Part #'),
        ),
        migrations.AlterField(
            model_name='assembly_item',
            name='comp',
            field=models.ForeignKey(help_text='The part number of the component of this item.', on_delete=django.db.models.deletion.CASCADE, related_name='assembly_item_comp', to='epic.Part', verbose_name='Component Part #'),
        ),
        migrations.AlterField(
            model_name='assembly_item',
            name='qty',
            field=models.PositiveIntegerField(help_text='The quantity of this component required by the assembly.'),
        ),
        migrations.AlterField(
            model_name='assembly_item',
            name='refdes',
            field=models.TextField(blank=True, default='', help_text='A list of comma-separated reference designators e.g., "R101,R304". The length of this list should match Quantity.', verbose_name='Refdes List'),
        ),
        migrations.AlterField(
            model_name='datasheet',
            name='ds_file',
            field=models.FileField(upload_to='epic/datasheets', verbose_name='Datasheet File'),
        ),
        migrations.AlterField(
            model_name='datasheet',
            name='notes',
            field=models.CharField(blank=True, help_text='For misc. information that may be useful along with the datasheet, such as website retrieved from.', max_length=1024),
        ),
        migrations.AlterField(
            model_name='delta',
            name='adj',
            field=models.IntegerField(help_text='The amount by which the part quantity should be adjusted by.', verbose_name='Adjustment Count'),
        ),
        migrations.AlterField(
            model_name='delta',
            name='is_absolute',
            field=models.BooleanField(default=False, help_text='If set, the adjustment is absolute otherwise it is relative.', verbose_name='Absolute Adjustment'),
        ),
        migrations.AlterField(
            model_name='delta',
            name='part',
            field=models.ForeignKey(help_text='The part whose quantity gets adjusted.', on_delete=django.db.models.deletion.CASCADE, to='epic.Part', verbose_name='Part #'),
        ),
        migrations.AlterField(
            model_name='delta',
            name='txtn',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='epic.Transaction', verbose_name='Transaction #'),
        ),
        migrations.AlterField(
            model_name='line_item',
            name='index',
            field=models.PositiveIntegerField(help_text='Sequential and unchanging index of this line-item.'),
        ),
        migrations.AlterField(
            model_name='line_item',
            name='line_cost',
            field=models.DecimalField(decimal_places=2, help_text='The total cost of this line-item.  Part-cost is line-item cost divided by quantity.', max_digits=9, validators=[django.core.validators.MinValueValidator(Decimal('0.00'))], verbose_name='Amount'),
        ),
        migrations.AlterField(
            model_name='line_item',
            name='part',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='epic.Part', verbose_name='Part #'),
        ),
        migrations.AlterField(
            model_name='line_item',
            name='qty',
            field=models.PositiveIntegerField(help_text='The quantity of this part.'),
        ),
        migrations.AlterField(
            model_name='line_item',
            name='txtn',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='epic.Transaction', verbose_name='Transaction #'),
        ),
        migrations.AlterField(
            model_name='order',
            name='expected_arrival_date',
            field=models.DateField(help_text='Date when the order is expected to arrive.', verbose_name='Expected Arrival Date'),
        ),
        migrations.AlterField(
            model_name='order',
            name='status',
            field=models.IntegerField(choices=[(0, 'open'), (1, 'closed')], default=0, verbose_name='Order Status'),
        ),
        migrations.AlterField(
            model_name='order',
            name='vendor',
            field=models.ForeignKey(help_text='The name of the vendor (distributor) where the order was placed.', on_delete=django.db.models.deletion.CASCADE, to='epic.Vendor'),
        ),
        migrations.AlterField(
            model_name='part',
            name='descr',
            field=models.CharField(blank=True, default='', help_text='Brief listing of key parameters of the part, such as temperature range, drift, max. voltages, etc.', max_length=127, verbose_name='Description'),
        ),
        migrations.AlterField(
            model_name='part',
            name='footprint',
            field=models.CharField(blank=True, default='', help_text="The part's PCB footprint.", max_length=63),
        ),
        migrations.AlterField(
            model_name='part',
            name='last_bom_mod_name',
            field=models.CharField(blank=True, default='', help_text='Name of entity which last modified this part.', max_length=31),
        ),
        migrations.AlterField(
            model_name='part',
            name='last_bom_mod_type',
            field=models.IntegerField(choices=[(0, 'user'), (1, 'tool')], default=0),
        ),
        migrations.AlterField(
            model_name='part',
            name='lead_time',
            field=models.PositiveIntegerField(help_text='Lead-time in weeks.', verbose_name='Lead-time [weeks]'),
        ),
        migrations.AlterField(
            model_name='part',
            name='mfg',
            field=models.CharField(help_text='The name of the manufacturer of the part.', max_length=31, verbose_name='Manufacturer'),
        ),
        migrations.AlterField(
            model_name='part',
            name='mfg_pn',
            field=models.CharField(max_length=31, verbose_name="Manufacturer's Part #"),
        ),
        migrations.AlterField(
            model_name='part',
            name='mounting',
            field=models.IntegerField(choices=[(0, 'SMD'), (1, 'THD'), (2, 'chassis'), (3, 'free')], default=0, help_text="How the part is mounted to a PCB (free if it's not mounted at all, such as a plug).", verbose_name='Mounting Type'),
        ),
        migrations.AlterField(
            model_name='part',
            name='overage',
            field=models.DecimalField(decimal_places=3, help_text='Worst-case percentage of parts we expect to lose due to attrition at the assembly-house.', max_digits=6, validators=[django.core.validators.MinValueValidator(Decimal('0')), django.core.validators.MaxValueValidator(Decimal('100'))], verbose_name='Overage [%]'),
        ),
        migrations.AlterField(
            model_name='part',
            name='spq',
            field=models.IntegerField(default=1, help_text='The number of parts in a standard package.  For example, 10,000 pieces in a reel.', verbose_name='Standard-Package Qty'),
        ),
        migrations.AlterField(
            model_name='part',
            name='status',
            field=models.IntegerField(choices=[(0, 'preview'), (1, 'active'), (2, 'deprecated'), (3, 'obsolete')], default=0, help_text="The life-time status of this part. Parts marked `preview' and `obsolete' are not considered orderable.", verbose_name='Life-time Status'),
        ),
        migrations.AlterField(
            model_name='part',
            name='substitute',
            field=models.ForeignKey(blank=True, help_text='List of other parts that are (identical) substitutes for this part.', null=True, on_delete=django.db.models.deletion.CASCADE, to='epic.Part', verbose_name='Substitutes'),
        ),
        migrations.AlterField(
            model_name='part',
            name='target_price',
            field=models.DecimalField(decimal_places=6, help_text='The expected price of the part.', max_digits=10, validators=[django.core.validators.MinValueValidator(Decimal('0.000001'))], verbose_name='Target price [$]'),
        ),
        migrations.AlterField(
            model_name='part',
            name='val',
            field=models.CharField(blank=True, default='', help_text='The primary value of the part such as resistance for a resistor or capacitance for a capacitor.', max_length=31, verbose_name='Value'),
        ),
        migrations.AlterField(
            model_name='shipment',
            name='cost_discount',
            field=models.DecimalField(decimal_places=2, help_text='Discounts given by the shipper, such as early payment discount.', max_digits=9, validators=[django.core.validators.MinValueValidator(Decimal('0.00'))], verbose_name='Discount Given'),
        ),
        migrations.AlterField(
            model_name='shipment',
            name='cost_freight',
            field=models.DecimalField(decimal_places=2, max_digits=9, validators=[django.core.validators.MinValueValidator(Decimal('0.00'))], verbose_name='Freight Cost'),
        ),
        migrations.AlterField(
            model_name='shipment',
            name='cost_other',
            field=models.DecimalField(decimal_places=2, help_text='Other costs assessed by the shipper, such as handling costs.', max_digits=9, validators=[django.core.validators.MinValueValidator(Decimal('0.00'))], verbose_name='Other Costs'),
        ),
        migrations.AlterField(
            model_name='shipment',
            name='from_warehouse',
            field=models.ForeignKey(blank=True, help_text='For an inter-warehouse shipment, the warehouse the shipment originates from.', null=True, on_delete=django.db.models.deletion.CASCADE, to='epic.Warehouse'),
        ),
        migrations.AlterField(
            model_name='shipment',
            name='ordr',
            field=models.ForeignKey(blank=True, help_text='For an order shipment, the order that resulted in this shipment.', null=True, on_delete=django.db.models.deletion.CASCADE, to='epic.Order', verbose_name='Order #'),
        ),
        migrations.AlterField(
            model_name='shipment',
            name='tracking',
            field=models.CharField(blank=True, default='', help_text='Comma-separated list of tracking numbers.', max_length=127, verbose_name='Tracking #s'),
        ),
        migrations.AlterField(
            model_name='transaction',
            name='notes',
            field=models.TextField(blank=True, help_text='Comments and notes for this transaction', verbose_name='Notes'),
        ),
        migrations.AlterField(
            model_name='transaction',
            name='ts',
            field=models.DateTimeField(help_text='Date and time when this transaction was created.', verbose_name='Creation Time'),
        ),
        migrations.AlterField(
            model_name='transaction',
            name='warehouse',
            field=models.ForeignKey(help_text='The (destination) warehouse this transaction applies to.', on_delete=django.db.models.deletion.CASCADE, to='epic.Warehouse', verbose_name='Warehouse'),
        ),
        migrations.AlterField(
            model_name='vendor',
            name='name',
            field=models.CharField(db_index=True, max_length=31, unique=True, verbose_name='Vendor Name'),
        ),
        migrations.AlterField(
            model_name='vendor',
            name='search_url',
            field=models.CharField(blank=True, default='', help_text="This pattern defines how to search for a particular part on the vendor's website.  %(vendor_pn)s gets replaced by the vendor's part-number, %(mfg)s by the manufacturer's name, and %(mfg_pn)s by the manufacturer's part-number.", max_length=127, verbose_name='Search URL Pattern'),
        ),
        migrations.AlterField(
            model_name='vendor_part',
            name='part',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='epic.Part', verbose_name='Part #'),
        ),
        migrations.AlterField(
            model_name='vendor_part',
            name='status',
            field=models.IntegerField(choices=[(0, 'preview'), (1, 'active'), (2, 'deprecated'), (3, 'obsolete')], default=1, help_text='The life-time status of this vendor part.', verbose_name='Life-time Status'),
        ),
        migrations.AlterField(
            model_name='vendor_part',
            name='vendor_pn',
            field=models.CharField(max_length=31, verbose_name="Vendor's Part #"),
        ),
        migrations.AlterField(
            model_name='warehouse',
            name='address',
            field=models.TextField(blank=True, default='', help_text='The shipping address for the warehouse.'),
        ),
        migrations.AlterField(
            model_name='warehouse',
            name='name',
            field=models.CharField(help_text='The name of the warehouse.', max_length=31, unique=True),
        ),
    ]



# Generated by Django 3.2.14 on 2022-07-25 04:13

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('epic', '0010_auto_20211108_1718'),
    ]

    operations = [
        migrations.AlterField(
            model_name='part',
            name='last_bom_mod_name',
            field=models.CharField(blank=True, default='', help_text='Name of entity which last modified this part.', max_length=127),
        ),
        migrations.AlterField(
            model_name='part',
            name='mfg',
            field=models.CharField(help_text='The name of the manufacturer of the part.', max_length=127, verbose_name='Manufacturer'),
        ),
        migrations.AlterField(
            model_name='part',
            name='mfg_pn',
            field=models.CharField(max_length=127, verbose_name="Manufacturer's Part #"),
        ),
        migrations.AlterField(
            model_name='part',
            name='val',
            field=models.CharField(blank=True, default='', help_text='The primary value of the part such as resistance for a resistor or capacitance for a capacitor.', max_length=127, verbose_name='Value'),
        ),
        migrations.AlterField(
            model_name='vendor',
            name='name',
            field=models.CharField(max_length=127, unique=True, verbose_name='Vendor Name'),
        ),
        migrations.AlterField(
            model_name='vendor_part',
            name='vendor_pn',
            field=models.CharField(max_length=127, verbose_name="Vendor's Part #"),
        ),
        migrations.AlterField(
            model_name='warehouse',
            name='name',
            field=models.CharField(help_text='The name of the warehouse.', max_length=127, unique=True),
        ),
    ]



#
#   Copyright (c) 2014-2015, 2017-2019 eGauge Systems LLC
# 	1644 Conestoga St, Suite 2
# 	Boulder, CO 80301
# 	voice: 720-545-9767
# 	email: davidm@egauge.net
#
#   All rights reserved.
#
#   This code is the property of eGauge Systems LLC and may not be
#   copied, modified, or disclosed without any prior and written
#   permission from eGauge Systems LLC.
#
"""Read footprint info from KiCad footprints directory."""

import os
import re

from django.conf import settings as cfg


class Footprints:
    @classmethod
    def get(cls):
        ret = []
        pattern = re.compile(r"([.]pretty(/|$))")
        for root, _, files in os.walk(cfg.EPIC_KICAD_FOOTPRINTS_DIR):
            libname = pattern.sub(
                ":", root[len(cfg.EPIC_KICAD_FOOTPRINTS_DIR) :]
            )
            for f in files:
                m = re.match(r"(.*)[.]kicad_mod$", f)
                if m:
                    ret.append(libname + m.group(1))
        return ret



# Generated by Django 5.2.4 on 2025-08-01 02:27

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('epic', '0012_alter_part_footprint'),
    ]

    operations = [
        migrations.AddIndex(
            model_name='assembly_item',
            index=models.Index(fields=['assy', 'comp'], name='epic_assemb_assy_id_dcc6cf_idx'),
        ),
        migrations.AddIndex(
            model_name='line_item',
            index=models.Index(fields=['txtn', 'part'], name='epic_line_i_txtn_id_bbdfa8_idx'),
        ),
        migrations.AddIndex(
            model_name='part',
            index=models.Index(fields=['mfg', 'mfg_pn'], name='epic_part_mfg_e03baa_idx'),
        ),
        migrations.AddIndex(
            model_name='line_item',
            index=models.Index(fields=['txtn', 'index'], name='epic_line_i_txtn_id_97b974_idx'),
        ),
    ]



# Generated by Django 3.2.14 on 2022-07-26 00:00

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('epic', '0011_auto_20220725_0413'),
    ]

    operations = [
        migrations.AlterField(
            model_name='part',
            name='footprint',
            field=models.CharField(blank=True, default='', help_text="The part's PCB footprint.", max_length=127),
        ),
    ]



#
#   Copyright (c) 2014-2015, 2018, 2021, 2024 eGauge Systems LLC
# 	1644 Conestoga St, Suite 2
# 	Boulder, CO 80301
# 	voice: 720-545-9767
# 	email: davidm@egauge.net
#
#   All rights reserved.
#
#   This code is the property of eGauge Systems LLC and may not be
#   copied, modified, or disclosed without any prior and written
#   permission from eGauge Systems LLC.
#
# pylint: disable=invalid-name
"""Unit tests.  Run, e.g., from epic-sample with:

python3 manage.py test epic.tests
"""

import io
import os
import re
from datetime import datetime, timedelta
from datetime import timezone as tz
from decimal import Decimal

from django.contrib.auth import get_user_model
from django.contrib.auth import models as auth_models
from django.test import TestCase
from django.urls import URLPattern, reverse
from django.utils import timezone

import epic
from epic import perms, urls
from epic.base import Enum
from epic.lib import get_stock
from epic.models import (
    Assembly_Item,
    Delta,
    Inventory,
    Line_Item,
    Order,
    Part,
    Shipment,
    Vendor,
    Vendor_Part,
    Warehouse,
)


def mktime(time_string):
    dt = datetime.strptime(time_string, "%b %d %Y %I:%M%p")
    return timezone.make_aware(dt, tz.utc)


class URL_Iterator:
    def __init__(self, urlpatterns):
        self.urlpattern_iterator = urlpatterns.__iter__()

    def __iter__(self):
        return self

    def __next__(self):
        while True:
            pat = self.urlpattern_iterator.__next__()
            if isinstance(pat, URLPattern):
                break
        regex = re.compile(pat.pattern._regex)  # NOQA
        args = {}
        for key in regex.groupindex.keys():
            args[key] = 1
        url = reverse("epic:%s" % pat.name, kwargs=args)
        return url, pat.name


def create_part(pk):
    return Part(
        pk=pk,
        mfg="mfg",
        mfg_pn="mfg_pn%d" % pk,
        mounting=Enum.MOUNTING_SMD,
        target_price=1,
        overage=0,
        spq=1,
        status=Enum.STATUS_ACTIVE,
        lead_time=9,
    )


def create_warehouse(pk, name):
    return Warehouse(pk=pk, name=name)


def create_vendor(pk, name):
    return Vendor(pk=pk, name=name)


def create_vendor_part(part_id, vendor_id, vendor_pn, price):
    return Vendor_Part(
        part_id=part_id, vendor_id=vendor_id, vendor_pn=vendor_pn, price=price
    )


def create_order(pk, vendor_id, warehouse_id):
    now = timezone.now()
    then = now + timedelta(days=30)
    return Order(
        pk=pk,
        ts=now,
        warehouse_id=warehouse_id,
        expected_arrival_date=then,
        status=Enum.TXTN_STATUS_OPEN,
        vendor_id=vendor_id,
    )


def create_ship(
    pk,
    order_id,
    warehouse_id,
    tracking="1Z9AA10123456784",
    cost_freight=0,
    cost_other=0,
    cost_discount=0,
):
    return Shipment(
        pk=pk,
        ts=timezone.now(),
        tracking=tracking,
        ordr_id=order_id,
        warehouse_id=warehouse_id,
        cost_freight=cost_freight,
        cost_other=cost_other,
        cost_discount=cost_discount,
    )


def create_inventory(pk, warehouse_id):
    return Inventory(pk=pk, ts=timezone.now(), warehouse_id=warehouse_id)


def create_assembly_item(assy_id=None, comp_id=None, qty=1):
    return Assembly_Item(assy_id=assy_id, comp_id=comp_id, qty=qty)


def create_line_item(txtn_id, part_id, qty, line_cost, index):
    return Line_Item(
        txtn_id=txtn_id,
        part_id=part_id,
        qty=qty,
        line_cost=line_cost,
        index=index,
    )


class Stock_Cache_Tests(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()

        cls.p = []  # parts
        cls.w = []  # warehouses
        cls.v = []  # vendors

        for i in range(10):
            p = create_part(i)
            p.save()
            cls.p.append(p)

        for i in range(3):
            w = create_warehouse(i, "Warehouse%d" % i)
            w.save()
            cls.w.append(w)

        for i in range(5):
            v = create_vendor(i, "Vendor%d" % i)
            v.save()
            cls.v.append(v)

    def assertInventory(self, inv, expected):
        for w in inv.warehouse:
            if w not in expected:
                self.fail("Warehouse %d unexpectedly in inventory." % w)
            else:
                for p in inv.warehouse[w]:
                    if p not in expected[w]:
                        self.fail(
                            "Warehouse %d: part %d unexpectedly in "
                            "inventory." % (w, p)
                        )
                    else:
                        if inv.warehouse[w][p].qty != expected[w][p]:
                            self.fail(
                                "Warehouse %d, part %d: wrong qty %d; "
                                "expected %d!"
                                % (
                                    w,
                                    p,
                                    inv.warehouse[w][p].qty,
                                    expected[w][p],
                                )
                            )

                for p in expected[w]:
                    if p not in inv.warehouse[w]:
                        self.fail(
                            "Warehouse %d: part %d missing from "
                            "inventory." % (w, p)
                        )

        for w in expected:
            if w not in inv.warehouse:
                self.fail("Warehouse %d missing from inventory." % w)

    def test_basic(self):
        inv0 = create_inventory(0, 0)
        inv0.ts = mktime("Jan 1 2010 4:00pm")
        inv0.save()
        inv0.finalize()

        d = Delta(
            part=self.__class__.p[0],
            is_absolute=True,
            adj=10,
            txtn=inv0,
            warehouse=self.__class__.w[0],
        )
        d.save()
        d = Delta(
            part=self.__class__.p[5],
            is_absolute=True,
            adj=20,
            txtn=inv0,
            warehouse=self.__class__.w[0],
        )
        d.save()
        d = Delta(
            part=self.__class__.p[3],
            is_absolute=True,
            adj=1000,
            txtn=inv0,
            warehouse=self.__class__.w[2],
        )
        d.save()

        inv = get_stock()
        self.assertInventory(inv, {0: {0: 10, 5: 20}, 2: {3: 1000}})

        inv = get_stock(warehouses=[0])
        self.assertInventory(inv, {0: {0: 10, 5: 20}})

        inv = get_stock(warehouses=[1])
        self.assertInventory(inv, {})

        inv = get_stock(warehouses=[2])
        self.assertInventory(inv, {2: {3: 1000}})

        inv = get_stock(exclude_txtn_id=inv0.id)
        self.assertInventory(inv, {})

        inv = get_stock(as_of_ts=mktime("Jan 1 2010 3:59pm"))
        self.assertInventory(inv, {})
        inv = get_stock(as_of_date=mktime("Jan 1 2010 5:00pm"))
        self.assertInventory(inv, {})

    def test_shipment_timestamp_change(self):
        """Test shipment that moves between timestamps."""
        o = create_order(pk=0, vendor_id=0, warehouse_id=0)
        o.save()
        o.finalize()
        s = create_ship(pk=1, order_id=o.id, warehouse_id=0)
        s.ts = mktime("Jan 1 2010 4:00pm")
        s.save()
        l = create_line_item(
            txtn_id=s.id, part_id=1, qty=2, line_cost="2", index=1
        )
        l.save()
        s.finalize()

        inv = get_stock(as_of_date=mktime("Jan 1 2015 5:00pm"))
        self.assertInventory(inv, {0: {1: 2}})
        inv = get_stock(as_of_date=mktime("Jan 1 2001 5:00pm"))
        self.assertInventory(inv, {})

        s.ts = mktime("Jan 1 2017 5:00pm")
        s.save()
        s.finalize()

        inv = get_stock(as_of_date=mktime("Jan 1 2015 5:00pm"))
        self.assertInventory(inv, {})
        inv = get_stock(as_of_date=mktime("Jan 1 2001 5:00pm"))
        self.assertInventory(inv, {})

        s.ts = mktime("Jan 1 2010 4:00pm")
        s.save()
        s.finalize()

        inv = get_stock(as_of_date=mktime("Jan 1 2015 5:00pm"))
        self.assertInventory(inv, {0: {1: 2}})
        inv = get_stock(as_of_date=mktime("Jan 1 2001 5:00pm"))
        self.assertInventory(inv, {})

    def test_inventory_timestamp_change(self):
        """Test inventory that moves between timestamps."""
        inv0 = create_inventory(0, 0)
        inv0.ts = mktime("Jan 1 2010 4:00pm")
        inv0.save()
        inv0.finalize()

        d = Delta(
            part=self.__class__.p[0],
            is_absolute=True,
            adj=10,
            txtn=inv0,
            warehouse=self.__class__.w[0],
        )
        d.save()
        d = Delta(
            part=self.__class__.p[5],
            is_absolute=True,
            adj=20,
            txtn=inv0,
            warehouse=self.__class__.w[0],
        )
        d.save()
        d = Delta(
            part=self.__class__.p[3],
            is_absolute=True,
            adj=1000,
            txtn=inv0,
            warehouse=self.__class__.w[2],
        )
        d.save()

        inv = get_stock(as_of_date=mktime("Jan 1 2015 5:00pm"))
        self.assertInventory(inv, {0: {0: 10, 5: 20}, 2: {3: 1000}})
        inv = get_stock(as_of_date=mktime("Jan 1 2001 5:00pm"))
        self.assertInventory(inv, {})

        inv0.ts = mktime("Jan 1 2017 5:00pm")
        inv0.save()
        inv0.finalize()

        inv = get_stock(as_of_date=mktime("Jan 1 2015 5:00pm"))
        self.assertInventory(inv, {})
        inv = get_stock(as_of_date=mktime("Jan 1 2001 5:00pm"))
        self.assertInventory(inv, {})

        inv0.ts = mktime("Jan 1 2010 4:00pm")
        inv0.save()
        inv0.finalize()

        inv = get_stock(as_of_date=mktime("Jan 1 2015 5:00pm"))
        self.assertInventory(inv, {0: {0: 10, 5: 20}, 2: {3: 1000}})
        inv = get_stock(as_of_date=mktime("Jan 1 2001 5:00pm"))
        self.assertInventory(inv, {})

    def test_warehouse_delete(self):
        """Test warehouse delete effect on stock cache."""
        item = create_assembly_item(assy_id=3, comp_id=1, qty=10)
        item.save()

        inv0 = create_inventory(pk=0, warehouse_id=2)
        inv0.ts = mktime("Jan 1 2010 4:00pm")
        inv0.save()
        inv0.finalize()

        d = Delta(
            part=self.__class__.p[0],
            is_absolute=True,
            adj=10,
            txtn=inv0,
            warehouse=self.__class__.w[0],
        )
        d.save()
        d = Delta(
            part=self.__class__.p[5],
            is_absolute=True,
            adj=20,
            txtn=inv0,
            warehouse=self.__class__.w[0],
        )
        d.save()
        d = Delta(
            part=self.__class__.p[3],
            is_absolute=True,
            adj=1000,
            txtn=inv0,
            warehouse=self.__class__.w[2],
        )
        d.save()

        inv = get_stock()
        self.assertInventory(inv, {0: {0: 10, 5: 20}, 2: {3: 1000}})

        Delta.objects.filter(pk__lte=2).delete()
        Warehouse.objects.filter(pk=0).delete()

        inv = get_stock()
        self.assertInventory(inv, {2: {3: 1000}})


class Part_Method_Tests(TestCase):
    def test_is_orderable(self):
        p = Part(status=Enum.STATUS_PREVIEW)
        self.assertEqual(p.is_orderable(), False)

        p = Part(status=Enum.STATUS_ACTIVE)
        self.assertEqual(p.is_orderable(), True)

        p = Part(status=Enum.STATUS_DEPRECATED)
        self.assertEqual(p.is_orderable(), True)

        p = Part(status=Enum.STATUS_OBSOLETE)
        self.assertEqual(p.is_orderable(), False)

    def test_equivalent_parts(self):
        p1 = create_part(1)
        p1.save()
        p2 = create_part(2)
        p2.save()

        self.assertEqual(p1.equivalent_parts(), [p1])

        p1.substitute = p2
        p1.save()
        p2.save()
        self.assertEqual(p1.equivalent_parts(), [p1, p2])
        self.assertEqual(p2.equivalent_parts(), [p1, p2])

        p3 = create_part(3)
        p3.substitute = p2
        p3.save()
        self.assertEqual(p1.equivalent_parts(), [p1, p2, p3])
        self.assertEqual(p2.equivalent_parts(), [p1, p2, p3])
        self.assertEqual(p3.equivalent_parts(), [p1, p2, p3])

        p1.substitute = None
        p1.save()
        self.assertEqual(p1.equivalent_parts(), [p1])
        self.assertEqual(p2.equivalent_parts(), [p2, p3])
        self.assertEqual(p3.equivalent_parts(), [p2, p3])

        p3.delete()
        p1.delete()
        p2.delete()

    def test_best_parts(self):
        p1 = create_part(1)
        p1.status = Enum.STATUS_PREFERRED
        p1.save()

        p2 = create_part(2)
        p2.status = Enum.STATUS_PREFERRED
        p2.substitute = p1
        p2.save()

        # p1 has lower ID so it should be preferred:
        self.assertEqual(p1.best_part(), p1)
        self.assertEqual(p2.best_part(), p1)

        p2.target_price = p1.target_price - 1
        p2.save()

        # p2 has lower price so it should be preferred:
        self.assertEqual(p1.best_part(), p2)
        self.assertEqual(p2.best_part(), p2)

        p2.status = Enum.STATUS_ACTIVE
        p2.save()
        # p1 is preferred, even though it's higher price
        self.assertEqual(p1.best_part(), p1)
        self.assertEqual(p2.best_part(), p1)

        p1.status = Enum.STATUS_ACTIVE
        p1.save()
        # now p2 should win again due to lower price:
        self.assertEqual(p1.best_part(), p2)
        self.assertEqual(p2.best_part(), p2)

    def test_avg_cost(self):
        part = []
        for i in range(3):
            p = create_part(i)
            if i == 1:
                p.substitute = part[0]
            p.save()
            part.append(p)

        for i in range(3):
            w = create_warehouse(i, "Warehouse%d" % i)
            w.save()

        for i in range(5):
            v = create_vendor(i, "Vendor%d" % i)
            v.save()

        # create shipment with:
        # 	part 1 cost average = $1
        # 	part 2 cost average = $0.001
        o = create_order(pk=0, vendor_id=0, warehouse_id=0)
        o.save()
        o.finalize()
        s = create_ship(pk=1, order_id=o.id, warehouse_id=0)
        s.save()
        l = create_line_item(
            txtn_id=s.id, part_id=0, qty=1, line_cost="3", index=1
        )
        l.save()
        l = create_line_item(
            txtn_id=s.id, part_id=1, qty=1, line_cost="1", index=2
        )
        l.save()
        l = create_line_item(
            txtn_id=s.id, part_id=2, qty=1000, line_cost="1", index=3
        )
        l.save()
        s.finalize()

        # Create shipment over a year ago.  It should not affect average prices
        o = create_order(pk=2, vendor_id=0, warehouse_id=0)
        o.save()
        o.finalize()
        s = create_ship(pk=3, order_id=o.id, warehouse_id=0)
        s.ts = timezone.now() - timedelta(days=365)
        s.save()
        l = create_line_item(
            txtn_id=s.id, part_id=1, qty=1, line_cost="1000", index=1
        )
        l.save()
        l = create_line_item(
            txtn_id=s.id, part_id=2, qty=1, line_cost="1000", index=2
        )
        l.save()
        s.finalize()

        p = Part.objects.get(pk=0)
        self.assertEqual(p.avg_cost(), Decimal("2.000"))

        p = Part.objects.get(pk=1)
        self.assertEqual(p.avg_cost(), Decimal("2.000"))

        p = Part.objects.get(pk=2)
        self.assertEqual(p.avg_cost(), Decimal("0.001"))


class Access_Tests(TestCase):
    @classmethod
    def setup_part(cls):
        p1 = create_part(1)
        p1.save()
        ai = create_assembly_item(assy_id=1, comp_id=2, qty=3)
        ai.refdes = "R1,R2,R3"
        ai.save()
        p = create_vendor_part(1, 2, "ASSY_HOUSE_VENDOR_PART_#1", 325.654321)
        p.save()

    @classmethod
    def setup_vendor(cls):
        v = create_vendor(1, "Vendor #1")
        v.save()
        p = create_vendor_part(1, 1, "VENDOR_PART_#1", 535.123456)
        p.save()

    @classmethod
    def setup_warehouse(cls):
        w = create_warehouse(1, "Warehouse #1")
        w.save()

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        perms.add_custom_permissions(auth_models)
        u = get_user_model().objects.create_user(
            "test", "test@magic.com", "magic"
        )
        u.user_permissions.set([perms.VIEW_PERM, perms.EDIT_PERM])

        p2 = create_part(2)
        p2.save()

        p3 = create_part(3)
        p3.save()

        w = create_warehouse(2, "Assyhouse #1")
        w.save()
        v = create_vendor(2, "Assyhouse #1")
        v.save()

        Access_Tests.setup_part()
        Access_Tests.setup_vendor()
        Access_Tests.setup_warehouse()

    # def test_redirect_to_https(self):
    #     url = reverse('epic:epic_index')
    #     response = self.client.get(url, secure=False)
    #     self.assertEqual(response.status_code, 301)
    #     self.assertIsNotNone(re.match('https://', response.url))

    def test_unauthenticated_accesses(self):
        response = self.client.get(
            reverse("epic:part_info") + "?pid=1", secure=True
        )
        self.assertEqual(response.status_code, 302)

        for url, _ in URL_Iterator(urls.urlpatterns):
            m = re.match(r"/epic/dal/", url)
            # auto-completion URLs respond with 403 (Forbidden)
            # others return redirect to login page:
            if m:
                expected = 403
            else:
                expected = 302
            response = self.client.get(url, secure=True)
            if response.status_code != expected:
                print(
                    "Unauthenticated access to %s returned %s, expected %s"
                    % (url, response.status_code, expected)
                )
            self.assertEqual(response.status_code, expected)

    def test_authenticated_accesses(self):
        self.client.login(username="test", password="magic")

        for url, pattern_name in URL_Iterator(urls.urlpatterns):
            if re.match("order_", pattern_name):
                # order part 1 (an assembly) from vendor 'Assyhouse #1'
                o = create_order(1, 2, 1)
                o.save()
                l = create_line_item(1, 1, qty=13, line_cost="15.95", index=1)
                l.save()
            elif re.match("ship_", pattern_name):
                # order part 1(an assembly) from vendor 'Assyhouse #1'
                o = create_order(2, 2, 1)
                o.save()
                l = create_line_item(2, 1, qty=13, line_cost="15.95", index=1)
                l.save()
                s = create_ship(1, 2, 1, "1234,5678", 1.1, 2.2, 3.3)
                s.save()
                l = create_line_item(1, 1, qty=2, line_cost="2", index=1)
                l.save()
            elif re.match("warehouse_inventory_", pattern_name):
                i = create_inventory(1, 1)
                i.save()

            expect = 200
            expect_url = None
            post_action = []
            if pattern_name == "search_results":
                url += "?q=1"
                # with a unique result, search redirects to the detail page:
                expect = 302
                expect_url = reverse("epic:part_detail", kwargs={"pk": 1})
            elif pattern_name == "part_delete":
                expect = 302
                expect_url = reverse("epic:part_list")
                post_action.append(Access_Tests.setup_part)
            elif pattern_name == "ship_delete":
                expect = 302
                expect_url = reverse("epic:ship_list")
            elif pattern_name == "vendor_delete":
                expect = 302
                expect_url = reverse("epic:vendor_list")
                post_action.append(Access_Tests.setup_vendor)
            elif pattern_name == "warehouse_delete":
                expect = 302
                expect_url = reverse("epic:warehouse_list")
                post_action.append(Access_Tests.setup_warehouse)
            elif pattern_name == "warehouse_inventory_delete":
                expect = 302
                expect_url = reverse("epic:warehouse_detail", kwargs={"pk": 1})
            elif pattern_name == "order_delete":
                expect = 302
                expect_url = reverse("epic:order_list")
            elif pattern_name in [
                "datasheet_add_part",
                "datasheet_detail",
                "datasheet_edit",
                "datasheet_delete",
            ]:
                continue

            response = self.client.get(url, secure=True)

            # we can't only have one transaction with pk=1...
            Shipment.objects.all().delete()
            Order.objects.all().delete()
            Inventory.objects.all().delete()

            if response.status_code != expect:
                print("access to %s failed" % url)
                if hasattr(response, "url"):
                    print(" redirect URL: %s" % response.url)
            self.assertEqual(response.status_code, expect)
            if expect_url is not None:
                url = response.url
                if url[-1] == "/":
                    url = url[:-1]
                self.assertEqual(url, expect_url)

            for action in post_action:
                action()

    def test_form_upload(self):
        self.client.login(username="test", password="magic")

        # Verify that we can post item formset to an order:
        response = self.client.post(
            "/epic/order/add",
            {
                "ts": "2018-08-06",
                "vendor": 1,
                "warehouse": 2,
                "expected_arrival_date": "2018-08-06",
                "notes": "This is a note",
                "line_item_set-TOTAL_FORMS": 12,
                "line_item_set-INITIAL_FORMS": 0,
                "line_item_set-MIN_NUM_FORMS": 0,
                "line_item_set-MAX_NUM_FORMS": 1000,
                "line_item_set-0-qty": 8,
                "line_item_set-0-part": 3,
                "line_item_set-0-vendor_pn": "VPN #1",
                "line_item_set-0-part_cost": 1,
                "line_item_set-0-line_cost": 120,
                "line_item_set-1-qty": 18,
                "line_item_set-1-part": 2,
                "line_item_set-1-vendor_pn": "VPN #2",
                "line_item_set-1-part_cost": 1,
                "line_item_set-1-line_cost": 121,
            },
        )

        # response.request['wsgi.input']._FakePayload__content.seek(0)
        # print('REQUEST',
        #       response.request['wsgi.input']._FakePayload__content.read()
        #       .decode('utf-8'))
        if response.status_code != 302:
            for name in response.context.keys():
                if re.match(r"[^_]+_form$", name):
                    print("%s.errors:" % name, response.context[name].errors)

        self.assertEqual(response.status_code, 302)
        order = Order.objects.all()[0]
        line_items = Line_Item.objects.filter(txtn=order)
        self.assertEqual(line_items[0].qty, 8)
        self.assertEqual(line_items[0].part.id, 3)
        self.assertEqual(line_items[0].line_cost, 120)
        self.assertEqual(line_items[1].qty, 18)
        self.assertEqual(line_items[1].part.id, 2)
        self.assertEqual(line_items[1].line_cost, 121)

        # Verify that we can post an item formset to a shipment:
        response = self.client.post(
            "/epic/order/%d/add-shipment" % order.id,
            {
                "ts": "2018-08-06",
                "cost_freight": 10.0,
                "cost_other": 3.14,
                "cost_discount": 2.74,
                "vendor": 1,
                "warehouse": 2,
                "expected_arrival_date": "2018-08-06",
                "notes": "note 2",
                "line_item_set-TOTAL_FORMS": 12,
                "line_item_set-INITIAL_FORMS": 0,
                "line_item_set-MIN_NUM_FORMS": 0,
                "line_item_set-MAX_NUM_FORMS": 1000,
                "line_item_set-0-qty": 7,
                "line_item_set-0-part": 3,
                "line_item_set-0-vendor_pn": "VPN #1",
                "line_item_set-0-part_cost": 1,
                "line_item_set-0-line_cost": 120,
                "line_item_set-1-qty": 17,
                "line_item_set-1-part": 2,
                "line_item_set-1-vendor_pn": "VPN #2",
                "line_item_set-1-part_cost": 1,
                "line_item_set-1-line_cost": 121,
            },
        )
        self.assertEqual(response.status_code, 302)
        ship = Shipment.objects.all()[0]
        line_items = Line_Item.objects.filter(txtn=ship)
        self.assertEqual(line_items[0].qty, 7)
        self.assertEqual(line_items[0].part.id, 3)
        self.assertEqual(line_items[0].line_cost, 120)
        self.assertEqual(line_items[1].qty, 17)
        self.assertEqual(line_items[1].part.id, 2)
        self.assertEqual(line_items[1].line_cost, 121)

        # Verify that we can post a delta formset as an inventory:
        response = self.client.post(
            "/epic/warehouse/1/add-inventory",
            {
                "ts": "2018-08-06 08:15 AM",
                "notes": "Note #3",
                "delta_set-TOTAL_FORMS": 12,
                "delta_set-INITIAL_FORMS": 0,
                "delta_set-MIN_NUM_FORMS": 0,
                "delta_set-MAX_NUM_FORMS": 1000,
                "delta_set-0-adj": 6,
                "delta_set-0-part": 2,
                "delta_set-0-vendor_pn": "VPN #2",
                "delta_set-0-part_cost": 1,
                "delta_set-0-line_cost": 120,
                "delta_set-1-adj": 16,
                "delta_set-1-part": 3,
                "delta_set-1-vendor_pn": "VPN #3",
                "delta_set-1-part_cost": 1,
                "delta_set-1-line_cost": 121,
            },
        )
        self.assertEqual(response.status_code, 302)
        inv = Inventory.objects.all()[0]
        line_items = Delta.objects.filter(txtn=inv)
        self.assertEqual(line_items[0].part.id, 2)
        self.assertEqual(line_items[0].is_absolute, True)
        self.assertEqual(line_items[0].adj, 6)
        self.assertEqual(line_items[0].warehouse.id, 1)
        self.assertEqual(line_items[1].part.id, 3)
        self.assertEqual(line_items[1].is_absolute, True)
        self.assertEqual(line_items[1].adj, 16)
        self.assertEqual(line_items[1].warehouse.id, 1)

    def test_spreadsheet_upload(self):
        self.client.login(username="test", password="magic")

        # Verify invalid column names yield the expected errors:
        response = self.client.post(
            "/epic/order/add",
            {
                "ts": "2018-08-06",
                "vendor": 1,
                "warehouse": 2,
                "expected_arrival_date": "2018-08-06",
                "notes": "This is a note",
                "spreadsheet": io.BytesIO(
                    b"qty1,part1,extra,amount1\n10,3,5,99\n40,EP0002,,10\n"
                ),
                "line_item_set-TOTAL_FORMS": 0,
                "line_item_set-INITIAL_FORMS": 0,
            },
        )
        self.assertContains(
            response,
            "<strong>Column &quot;qty&quot; missing.  It may also be "
            "called one of: quantity, count.</strong></span> "
            '<span id="error_2_id_spreadsheet" class="help-block">'
            "<strong>Column &quot;part&quot; missing.  It may also be "
            "called one of: pn, ep.</strong></span> "
            '<span id="error_3_id_spreadsheet" class="help-block">'
            "<strong>Column &quot;amount&quot; missing.  It may also be "
            "called one of: cost, price.</strong>",
        )

        # Verify that we can post a CSV file to an order:
        response = self.client.post(
            "/epic/order/add",
            {
                "ts": "2018-08-06",
                "vendor": 1,
                "warehouse": 2,
                "expected_arrival_date": "2018-08-06",
                "notes": "This is a note",
                "spreadsheet": io.BytesIO(
                    b"qty,part,extra,amount\n10,3,5,99\n40,EP0002,,10\n"
                ),
                "line_item_set-TOTAL_FORMS": 0,
                "line_item_set-INITIAL_FORMS": 0,
            },
        )
        self.assertEqual(response.status_code, 302)
        order = Order.objects.all()[0]
        line_items = Line_Item.objects.filter(txtn=order)
        self.assertEqual(line_items[0].qty, 10)
        self.assertEqual(line_items[0].part.id, 3)
        self.assertEqual(line_items[0].line_cost, 99)
        self.assertEqual(line_items[1].qty, 40)
        self.assertEqual(line_items[1].part.id, 2)
        self.assertEqual(line_items[1].line_cost, 10)

        # Verify that we can post a CSV file to a shipment:
        response = self.client.post(
            "/epic/order/%d/add-shipment" % order.id,
            {
                "ts": "2018-08-06",
                "cost_freight": 10.0,
                "cost_other": 3.14,
                "cost_discount": 2.74,
                "vendor": 1,
                "warehouse": 2,
                "expected_arrival_date": "2018-08-06",
                "notes": "note 2",
                "spreadsheet": io.BytesIO(
                    b"qty,part,extra,amount\n5,3,5,99\n10,EP0002,,10\n"
                ),
            },
        )
        self.assertEqual(response.status_code, 302)
        ship = Shipment.objects.all()[0]
        line_items = Line_Item.objects.filter(txtn=ship)
        self.assertEqual(line_items[0].qty, 5)
        self.assertEqual(line_items[0].part.id, 3)
        self.assertEqual(line_items[0].line_cost, 99)
        self.assertEqual(line_items[1].qty, 10)
        self.assertEqual(line_items[1].part.id, 2)
        self.assertEqual(line_items[1].line_cost, 10)

        # Verify that we can post an Excel file as an inventory:
        excel_file = os.path.join(
            os.path.dirname(epic.__file__), "test-data", "inventory.xlsx"
        )
        with open(excel_file, "rb") as f:
            response = self.client.post(
                "/epic/warehouse/1/add-inventory",
                {"ts": "2018-08-06 08:15 AM", "spreadsheet": f},
            )
        self.assertEqual(response.status_code, 302)
        inv = Inventory.objects.all()[0]
        line_items = Delta.objects.filter(txtn=inv)
        self.assertEqual(line_items[0].part.id, 2)
        self.assertEqual(line_items[0].is_absolute, True)
        self.assertEqual(line_items[0].adj, 1234)
        self.assertEqual(line_items[0].warehouse.id, 1)
        self.assertEqual(line_items[1].part.id, 3)
        self.assertEqual(line_items[1].is_absolute, True)
        self.assertEqual(line_items[1].adj, 5678)
        self.assertEqual(line_items[1].warehouse.id, 1)

    def test_order_status(self):
        """Confirm that deleting a shipment for completed order will mark
        order as OPEN again.

        """
        self.client.login(username="test", password="magic")
        # create order from Vendor #1:
        o = create_order(pk=4, vendor_id=1, warehouse_id=1)
        o.save()
        l = create_line_item(o.pk, 2, qty=1000, line_cost="256.35", index=1)
        l.save()

        # order should be open now
        self.assertEqual(o.status, Enum.TXTN_STATUS_OPEN)

        # create two shipments that should complete the order:
        s = create_ship(2, o.pk, 1, "1Z1345", 1.23, 2.34, 3.45)
        s.save()
        l = create_line_item(s.pk, 2, 900, line_cost="230.05", index=1)
        l.save()

        s = create_ship(3, o.pk, 1, "1Z1356", 1.23, 2.34, 3.45)
        s.save()
        l = create_line_item(s.pk, 2, 100, line_cost="230.05", index=1)
        l.save()

        s.finalize()

        # order should be closed now:
        o = Order.objects.get(pk=4)
        self.assertEqual(o.status, Enum.TXTN_STATUS_CLOSED)

        # delete the first shipment:
        url = reverse("epic:ship_delete", kwargs={"pk": 2})
        response = self.client.get(url, secure=True)
        self.assertEqual(response.status_code, 302)

        # order should be open again:
        o = Order.objects.get(pk=4)
        self.assertEqual(o.status, Enum.TXTN_STATUS_OPEN)



#
#   Copyright (c) 2021 eGauge Systems LLC
# 	1644 Conestoga St, Suite 2
# 	Boulder, CO 80301
# 	voice: 720-545-9767
# 	email: davidm@egauge.net
#
#   All rights reserved.
#
#   This code is the property of eGauge Systems LLC and may not be
#   copied, modified, or disclosed without any prior and written
#   permission from eGauge Systems LLC.
#
from django.urls import include, path
from rest_framework import routers

from . import views

router = routers.SimpleRouter()
router.register(r"part", views.PartViewSet, basename="part")
router.register(r"vendor", views.VendorViewSet, basename="vendor")
router.register(
    r"vendor_part", views.VendorPartViewSet, basename="vendor_part"
)
router.register(
    r"assembly_item", views.AssemblyItemViewSet, basename="assembly_view"
)

urlpatterns = [path("", include(router.urls))]



#
#   Copyright (c) 2021, 2025 eGauge Systems LLC
# 	4805 Sterling Dr, Suite 1
# 	Boulder, CO 80301
# 	voice: 720-545-9767
# 	email: davidm@egauge.net
#
#   All rights reserved.
#
#   This code is the property of eGauge Systems LLC and may not be
#   copied, modified, or disclosed without any prior and written
#   permission from eGauge Systems LLC.
#
"""This module implements a static cache of EPIC parts."""

from types import SimpleNamespace

from egauge import webapi

from .epic_client import EPICAPIClient
from .error import Error


def _append_range(ranges, start, end):
    """If START is None, do nothing.  Otherwise, append a new range to
    RANGES.  If END > START, the range "START-END" is added, with
    START and END being decimal integers.  If END == START, the
    decimal "START" is added to RANGES.

    """
    if start is not None:
        r = "%d" % start
        if end > start:
            r += "-%d" % end
        ranges.append(r)


def _range_list_from_set(s):
    """Return a string that concisely represents the set S as a list for
    ranges.  Each range is either a decimal number or two decimal
    numbers separated by '-'.  The ranges themselves are separated by
    commas.

    """
    l = sorted(list(s))
    ranges = []
    start = end = None
    for n in l:
        if start is None or end is None:
            start = end = n
        elif n == end + 1:
            end = n
        else:
            _append_range(ranges, start, end)
            start = end = n
    _append_range(ranges, start, end)
    return ",".join(ranges)


class PartsCache:
    """A static cache of EPIC parts."""

    needed = set()
    cache: dict[int, SimpleNamespace] = {}
    epic_api = None

    @classmethod
    def set_epic(cls, epic_api: EPICAPIClient):
        """Establishes EPIC_API as the EPIC API to use for fetching parts
        info.  This must be called prior to the first call of
        PartsCache.get().  If the EPIC API used changes, the cache is
        cleared.

        """
        if cls.epic_api != epic_api:
            cls.cache = {}
            cls.epic_api = epic_api

    @classmethod
    def prefetch(cls, part_number: int):
        """Prefetch the part with number PART_NUMBER.  This current
        implementation simply records the part number internall and
        does not actually cause any calls to the EPIC API.

        """
        if part_number in cls.cache or part_number in cls.needed:
            return
        cls.needed.add(part_number)

    @classmethod
    def get(cls, part_number: int):
        """Get the part with number PART_NUMBER.  If the part is already in
        the cache, the part will be returned immediately.  Otherwise,
        the part will be fetched through the EPIC API, along with any
        other parts that have been prefetched since the most recent
        call to thos method.  Once that info is received, the cache is
        updated and the requested part returned.

        """
        if cls.epic_api is None:
            raise Error("must call Parts.set_epic() first")

        if part_number not in cls.cache:
            if part_number not in cls.needed:
                cls.prefetch(part_number)
            r = _range_list_from_set(cls.needed)
            try:
                reply = cls.epic_api.get("part/?id=%s" % r)
            except webapi.Error as e:
                raise Error("Failed to get part.", r) from e
            if not isinstance(reply, list):
                raise Error("Invalid server response while getting parts.", r)
            for p in reply:
                part = SimpleNamespace(**p)
                cls.cache[part.id] = part
        return cls.cache[part_number]



#
#   Copyright (c) 2021, 2025 eGauge Systems LLC
# 	4805 Sterling Dr, Suite 1
# 	Boulder, CO 80301
# 	voice: 720-545-9767
# 	email: davidm@egauge.net
#
#   All rights reserved.
#
#   This code is the property of eGauge Systems LLC and may not be
#   copied, modified, or disclosed without any prior and written
#   permission from eGauge Systems LLC.
#
"""This module provides helper functions outputting the BOM as a CSV
file."""

from dataclasses import dataclass, fields
from pathlib import Path
from types import SimpleNamespace
from typing import Mapping, Sequence

import tablib

from epic.base import Enum, format_part_number

from .bom import BOM, Component
from .parts_cache import PartsCache
from .vendors_cache import VendorParts, VendorsCache


def _is_orderable(vp: SimpleNamespace) -> bool:
    """A vendor part is orderable only if itself and the corresponding
    part are orderable.

    Positional arguments:

    vp -- The vendor part to check whether its orderable.

    """
    if vp.status not in Enum.STATUS_ORDERABLE:
        return False
    part = PartsCache.get(vp.part)
    if part.status not in Enum.STATUS_ORDERABLE:
        return False
    return True


@dataclass
class BOMRow:
    qty: int | str | None = None
    value: str | None = None
    footprint: str | None = None
    pn: str | None = None
    mfg: str | None = None
    mfg_pn: str | None = None
    substitutes: str | None = None
    refdes: str | None = None  # colon-separated list of reference designators
    vendor_pns: list[str] | None = None

    def as_list(self) -> list:
        ret = []
        for f in fields(self.__class__):
            v = getattr(self, f.name)
            if isinstance(v, list):
                ret += v
            else:
                ret.append(v or "")
        return ret


class CSVBOM:
    def __init__(
        self,
        bom: BOM,
        vendors: None | Sequence[SimpleNamespace] = None,
    ):
        """Create a CSV version of the BOM.

        Positional arguments:

        bom -- The BOM to convert to CSV.

        vendors -- The list of vendors whose vendor part info to
                include in the CSV.  If `None`, no vendor part info
                at all is included.  If an empty list, vendor part info
                is included for all vendors that can supply at least one
                part for this BOM.  If a list of vendors, vendor part info
                for all of those vendors is included.

        """
        if vendors is None:
            self.vendors = tuple()
        elif vendors:
            self.vendors = vendors  # use select list of vendors
        else:
            # use all vendors that at least one part for this BOM:
            self.vendors = tuple(VendorsCache())

        self.vendor_parts = VendorParts(list(bom.comps.keys()), self.vendors)

        if vendors is not None and len(vendors) == 0:
            # only keep vendors which have at least one part for this BOM:
            self.vendors = tuple(
                v
                for v in VendorsCache()
                if self.vendor_parts.vendor_has_parts(v)
            )

        self.dataset = tablib.Dataset()
        self.dataset.headers = [
            "Qty",
            "Value",
            "Footprint",
            f"{bom.manufacturer} PN",
            "Mfg",
            "Mfg PN",
            "Approved Substitutes",
            "Refdes",
        ] + [f"{v.name} PN" for v in self.vendors]

        comp_list = self._bom_sort(bom.comps)
        self._output_list(comp_list)

        dnp_list = self._bom_sort(bom.dnps)
        if dnp_list:
            self.dataset.append(
                BOMRow(
                    qty="DO NOT PLACE parts:",
                    vendor_pns=len(self.vendors) * [""],
                ).as_list()
            )
            self._output_list(dnp_list)

    def save(self, filename: Path):
        """Save the CSV BOM in a file.

        Positional arguments:

        filename -- The path in which to save the CSV BOM.

        """
        with open(filename, "w") as f:
            f.write(self.dataset.export("csv"))

    def _bom_sort(
        self, comps: Mapping[int, Sequence[Component]]
    ) -> Sequence[Sequence[Component]]:
        return sorted(
            [comps for comps in comps.values() if comps],
            key=lambda comps: comps[0].part_id,
        )

    def _best_vendor_pn(
        self, part: SimpleNamespace, vendor: SimpleNamespace
    ) -> str:
        best_vp = None
        for pn in part.equivalents:
            vp = self.vendor_parts.get(pn, vendor)
            if not vp:
                continue

            if best_vp is None:
                best_vp = vp
                continue

            best_orderable = _is_orderable(best_vp)
            vp_orderable = _is_orderable(vp)

            if vp_orderable and (
                not best_orderable or vp.price < best_vp.price
            ):
                best_vp = vp
        if not best_vp:
            return ""
        return best_vp.vendor_pn

    def _output_list(self, bom_list: Sequence[Sequence[Component]]):
        """Append the components in a BOM list to the CSV BOM.

        Positional arguments:

        bom_list -- The list of components to append to the CSV.

        """
        for components in bom_list:
            c = components[0]
            if c.part:
                for sub_id in c.part.equivalents:
                    PartsCache.prefetch(sub_id)

        for components in bom_list:
            qty = len(components)
            refdes = ":".join(sorted([c.refdes for c in components]))
            c = components[0]

            substitutes = []
            vendor_pns = len(self.vendors) * [""]
            if c.part:
                for sub_id in c.part.equivalents:
                    if sub_id == c.part_id:
                        continue
                    sub = PartsCache.get(sub_id)
                    substitutes.append(
                        f"{format_part_number(sub.id)} ({sub.mfg} {sub.mfg_pn})"
                    )

                    vendor_pns = [
                        self._best_vendor_pn(c.part, v) for v in self.vendors
                    ]

            row = BOMRow(
                qty=qty,
                value=c.value,
                footprint=c.footprint,
                pn=format_part_number(c.part_id),
                mfg=c.mfg,
                mfg_pn=c.mfg_pn,
                substitutes=":".join(substitutes),
                refdes=refdes,
                vendor_pns=vendor_pns,
            )
            self.dataset.append(row.as_list())
